--[[
Copyright (C) 2016-2017  Neemys

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

See <LICENSE.txt> for a copy of this license, or see <http://www.gnu.org/licenses/>
]]


require("misc")
require("constants")
require("config")

--will store platform tile to layer relation to check in collsion_mask if a building has been restricted with by a specific tile. As platforms do not block each other buildings using collision mask, it will help checking if a building is blocked by a platform using script check
platformToLayer = nil


--Using the dummy entities that we created in data-final-fixes phase, it will populate the platformToLayer variable for the script to use.
local function fillPlatformToLayer(game)
	platformToLayer = {}
	for _,tileProto in pairs(game.tile_prototypes) do
		if(game.entity_prototypes[buildingplatform.constants.dummyPrefix..tileProto.name]) then
			platformToLayer[tileProto.name] = game.entity_prototypes[buildingplatform.constants.dummyPrefix..tileProto.name].order
		end
	end
end

-- create the initial surface list when it does not exist.
local function createSurfaceList()
	global.surfaceList={nauvis=true} -- The default surface accept platform by default
end


--[[API_DOC
@interface building-platform
@description Check that the given placed entity is blocked by ground or tile under it by rules set at game loading, but do not actually remove the entity.

If this function return false and you don't remove the building, it can disappear by itself under certain circumstance. Like if tile under it is removed because the game's collision_mask check might see a collision with the now uncovered ground. When the game sees a collision like that, it make the building disappear without firing a destruction event or error message. It even don't give the building back to the player. It just vanishes.
@param entity table Entity which placement need to be checked.
@return False if the entity is blocked, true otherwise.
]]--
checkGoodBuildingPlacement = function(entity)
	if not entity then return false end
	if not platformToLayer then fillPlatformToLayer(game) end --as we don't have access to game in control nor on_load, we have no choice but generate it here if needed
	local box = entity.bounding_box
	local isGhost = (entity.name == "entity-ghost");
	local entityPrototype = (isGhost and entity.ghost_prototype or entity.prototype );
	local collisionMask = entityPrototype.collision_mask or {}
	local isRestrictedBuilding = false
	--check if it is a restricted buildings on any platform
	for _,layer in pairs(platformToLayer) do
		if(collisionMask[layer]) then
			isRestrictedBuilding = true
		end
	end
	if isRestrictedBuilding then--obviously we only need to check restricted building
		for x=box.left_top.x,box.right_bottom.x,1 do
			for y=box.left_top.y,box.right_bottom.y,1 do
				local tilePrototype = entity.surface.get_tile(x,y).prototype
				local hiddenTileName = entity.surface.get_hidden_tile({x,y}) --get_hidden_tile return only the name, not the tile
				local hiddenTilePrototype
				if(hiddenTileName) then
					hiddenTilePrototype = game.tile_prototypes[hiddenTileName]
				end

				local visiblePlatformLayer = platformToLayer[tilePrototype.name]
				 --building blocked by visible platform
				if(visiblePlatformLayer and not collisionMask[visiblePlatformLayer]) then
					return false
				else--either the visible tile is not a platform or is a platform that allow the building
					local hiddenPlatformLayer = hiddenTilePrototype and platformToLayer[hiddenTilePrototype.name] or nil
					--in both case, if the hidden tile is a platform that block the building, it's blocked (platform over platform need the building to be allowed by both)
					if(hiddenPlatformLayer and not collisionMask[hiddenPlatformLayer]) then
						return false
					else--either the hidden tile is not a platform or is a platform that allow the building
						--we now have 4 case left to check, only one should block the building : if there is no platform in visible or hidden tile
						if(not visiblePlatformLayer and not hiddenPlatformLayer) then
							return false
						end
					end
				end
			end
		end
	end
	return true
end

--[[API_DOC
@interface building-platform
@description Destroy a given entity and give back to the player the item. The entity is looted in case it contains items.
@param entity table Entity to remove.
@param player table Player to give back item and loots.
@param needMessage boolean If set to true a default error message will be displayed to the player. Set to false if you want to send your own message.
@return True if everything went fine, false otherwise.
]]--
removeAndReimburseBuilding = function(entity,player,needMessage)
	if not entity then return false end
	if not player then return false end
	if not needMessage then needMessage = true end
	local isGhost = (entity.name == "entity-ghost");
	local entityPrototype = (isGhost and entity.ghost_prototype or entity.prototype );
	-- when removing, we need to loot it in case another mod put item inside
	-- need a solution if some item can't go inside inventory, maybe spawn a chest ? For the moment they just vanish
	if(not isGhost) then
		local loot = entityPrototype.loot
		if loot then
			for i,loot_ele in pairs(loot) do
				player.insert({name=loot.item,count=loot.count_max})
			end
		end
		for i,item in pairs(entity.prototype.mineable_properties.products) do
			player.insert({name=item.name})  -- The entity by itself is not included in loot list
		end
	end
	entity.destroy()
	--Tell the player why the building has been removed if asked to do it.
	if(needMessage) then
		player.print("This building cannot be build here !")
	end
	return true
end


--[[API_DOC
@interface building-platform
@description Check that the given placed entity is blocked by ground or tile under it under rules set at game loading.
If the entity is wrongly placed, remove it and give back to the player the item. The entity is looted in case it contains items.
Then show a message to the player.
Nothing is done if the building is placed correctly.
@param entity table Entity to check.
@param player table Player to give back item and loots.
@return False if the entity was blocked, true otherwise.
]]--
checkBuildingPlacementAndReimburseWithMessage = function(entity,player)
	local isGoodPlacement = checkGoodBuildingPlacement(entity)
	if(not isGoodPlacement) then
		removeAndReimburseBuilding(entity,player,true)
	end
	return isGoodPlacement
end

--[[API_DOC
@interface building-platform
@description Add the provided surface(s) name(s) to the list of allowed surfaces for the custom generation making them spawn platforms.
@param surfaces (string | table of strings) Surface(s) to add.
@return False if there was a problem with adding to the list, true otherwise.
]]--
addSurfacesToCustomPlacementList = function(surfaces)
	if not global.surfaceList then
		createSurfaceList()
	end
	if (type(surfaces) == "string") then
		surfaces = {surfaces}
	end
	for _,surface in pairs(surfaces) do
		global.surfaceList[surface] = true
	end
	return true
end

--[[API_DOC
@interface building-platform
@description Remove the provided surface(s) name(s) of the list of allowed surfaces for the custom generation making them not spawn platforms anymore.
@param surfaces (string | table of strings) Surface(s) to remove.
@return False if there was a problem with removing from the list, true otherwise.
]]--
removeSurfacesToCustomPlacementList = function(surfaces)
	if not global.surfaceList then
		createSurfaceList()
	end
	if (type(surfaces) == "string") then
		surfaces = {surfaces}
	end
	for _,surface in pairs(surfaces) do
		if(global.surfaceList[surface]) then
			global.surfaceList[surface] = nil
		end
	end
	return true
end

--[[API_DOC
@interface building-platform
@description Get the list of allowed surfaces for custom generation. This list is a table in the form [surface]=>boolean
@return The list of allowed surfaces.
]]--
getCustomPlacementAllowedSurfacesList = function()
	if not global.surfaceList then
		createSurfaceList()
	end
	return global.surfaceList
end


--Expose our functions in the interface.
remote.add_interface("building-platform",{
	checkGoodBuildingPlacement=checkGoodBuildingPlacement,
	removeAndReimburseBuilding=removeAndReimburseBuilding,
	checkBuildingPlacementAndReimburseWithMessage=checkBuildingPlacementAndReimburseWithMessage,
	addSurfacesToCustomPlacementList=addSurfacesToCustomPlacementList,
	removeSurfacesToCustomPlacementList=removeSurfacesToCustomPlacementList,
	getCustomPlacementAllowedSurfacesList=getCustomPlacementAllowedSurfacesList
})



-- When player build something, check if it is a restricted building and build on a platform. Remove it if it must be.
script.on_event(defines.events.on_built_entity, function(event)
	checkBuildingPlacementAndReimburseWithMessage(event.created_entity,game.players[event.player_index])
end)


-- In case a recipe is activated midgame and the research is already done, seek where the recipe is unlocked, check if the research is done and unlock the recipe. This function is ressource heavy, shouldn't be called often.
-- Can be used to hide the recipe if the recipe have been deactivated in config.
script.on_event(buildingplatform.constants.prefix.."reload-recipe", function(event)
	for _,force in pairs(game.forces) do
		if force.recipes[buildingplatform.constants.tileName] then -- only if the recipe exist
			if(settings.startup[buildingplatform.constants.craftableSetting].value) then
				local technologyFound = false
				local recipeToUnlocked = false
				local function inEffect(recipeName,effectsArray)
					for _,effect in pairs(effectsArray) do
						if effect['type'] == "unlock-recipe" and effect['recipe'] == recipeName then
							return true
						end
					end
					return false
				end
				--check for building platform default config
				if(force.technologies["concrete"] and inEffect(buildingplatform.constants.tileName,force.technologies["concrete"].effects)) then
					technologyFound = true
					if force.technologies["concrete"].researched then
						recipeToUnlocked = true
					end
				end
				--if not in default config, search where it has been put
				if not technologyFound then
					for _,tech in pairs(force.technologies) do
						if inEffect(buildingplatform.constants.tileName,tech.effects) then
							technologyFound = true
							if tech.researched then
								recipeToUnlocked = true
								break --stop the loop only if the recipe is to be unlocked, if not, maybe another tech that unlock the platform is researched
							end
						end
					end
				end
				if recipeToUnlocked then
					force.recipes[buildingplatform.constants.tileName].enabled = true
				end
			else --no craftable
				force.recipes[buildingplatform.constants.tileName].enabled = false --in case someone want to deactivate the recipe²²
			end
		end
	end
end)


-- custom platform generation code
if((settings.startup[buildingplatform.constants.platformGeneratorSetting].value == "concentric-circle" or settings.startup[buildingplatform.constants.platformGeneratorSetting].value == "grid" or settings.startup[buildingplatform.constants.platformGeneratorSetting].value == "default") and not buildingplatform.config.disableOwnPlatform) then


	-- functionToUseForGeneration should, for given coordinate x and y, return the name of the tile to change at that location, nil if no tile should be changed.
	local functionToUseForGeneration = function() end
	if(settings.startup[buildingplatform.constants.platformGeneratorSetting].value == "concentric-circle") then
		-- m2cengine's concentric-circle platform generator code
		functionToUseForGeneration = function ( x, y,surface )
			local ret = false
			local ret2 = false

			local function limit( a, b, v )
				return math.min( b, math.max( a, v ) )
			end

			local abs = math.abs
			local min = math.min
			local max = math.max
			local pow = math.pow
			local rand = math.random --factorio's math.random is deterministic so no desync in multi (https://forums.factorio.com/viewtopic.php?p=46431#p46431)
			local floor = math.floor

			r = math.sqrt( x * x + y * y )
			if math.floor( r ) % 150 < limit( 3, 16, 3 + .15 * r * .3 ) and rand() < .85 then
				ret = true
			end
			rmd = r * .01 * 1.0
			if rmd < 1 and rand() < 1.1 - pow( rmd, 4 ) then
				ret = true
			end
			if abs( x ) < 2 + 8 - .005 * abs( y ) and rand() < pow( 1 - .003 * r, .05 ) then ret = true end
			if abs( y ) < 2 + 8 - .005 * abs( x ) and rand() < pow( 1 - .003 * r, .05 ) then ret = true end

			if abs( x ) < 3 or abs( y ) < 3 then
				ret2 = true
			end

			if abs( x - y ) < 3 + .0 * r or abs( y + x ) < 3 + .0 * r then
				ret2 = true
			end

			if ( max( abs( x ), abs( y ) ) + 50 ) % 130 < 3 then
				ret2 = true
			end

			if rmd < .2 then
				ret = false
				ret2 = false
			end
			local tileNameToReturn = nil
			if ret == true then
				tileNameToReturn = buildingplatform.constants.tileName
			end
			if ret2 == true then
				tileNameToReturn = buildingplatform.constants.speedTileName
			end
			return tileNameToReturn
		end
	elseif(settings.startup[buildingplatform.constants.platformGeneratorSetting].value == "grid") then
		--we want a grid of square centered on 0,0 with a set side length and spacing.
		functionToUseForGeneration = function ( x, y,surface )
			local isPlatform = false
			local squareLength = settings.startup[buildingplatform.constants.sideLengthSetting].value
			local spacing = settings.startup[buildingplatform.constants.spacingBetweenPlatformSetting].value
			local halfSquare = math.ceil(squareLength/2) --used as an offset to center the square on 0,0

			if((x+halfSquare)%(squareLength+spacing) < squareLength and (y+halfSquare)%(squareLength+spacing) < squareLength) then
				isPlatform = true
			end

			if isPlatform then
				return buildingplatform.constants.tileName
			end
			return nil
		end
	elseif(settings.startup[buildingplatform.constants.platformGeneratorSetting].value == "default") then
		--we want to replace dummy platform resource by our platform tile
		functionToUseForGeneration = function ( x, y, surface )
			--log(serpent.block(surface.find_entities({{x,y},{x+1,y+1}})))
			local entitiesOnTile = surface.find_entities_filtered{area = {{x,y},{x+1,y+1}},name=buildingplatform.constants.dummyPlatformName}
			if(#entitiesOnTile == 1) then
				entitiesOnTile[1].destroy() -- no need to keep the dummy resource
				return buildingplatform.constants.tileName
			end
			return nil
		end
	end

	-- take care to apply the given function to a whole chunk with added check like surface name and if we do not replace water.
	generateChunkClosure = function(checkTileFunction)
		local checkTileFunction = checkTileFunction

		return function(event)
			local surfaceList = getCustomPlacementAllowedSurfacesList()
			if not surfaceList[event.surface.name] then --we only generate the platform for the allowed surfaces
				return
			end

			local x1 = event.area.left_top.x
			local y1 = event.area.left_top.y
			local x2 = event.area.right_bottom.x
			local y2 = event.area.right_bottom.y
			local surface = event.surface

			tiles = {}
			for x = x1, x2  do
				for y = y1, y2 do
					local old = surface.get_tile(x, y).name
					new = nil

					if not (old == "water" or old == "deepwater") then --we do not place our tile on water
						new = checkTileFunction( x, y , surface)
						if not (new == old) and new then
							table.insert(tiles, {name = new, position = {x, y}})
						end
					end
				end
			end
			surface.set_tiles(tiles)
		end
	end

	script.on_event(defines.events.on_chunk_generated, generateChunkClosure(functionToUseForGeneration))
end
