--[[
Copyright (C) 2016-2017  Neemys

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

See <LICENSE.txt> for a copy of this license, or see <http://www.gnu.org/licenses/>
]]

require('internals_functions')


--[[API_DOC
@description Block all buildings of a given type outside the given tileObject.

Calling this function will not remove previous permission.
It mean that you can allow a type on multiple tileObject by calling this function with the same type, but with each different tileObject you want.
@param buildingType string Factorio type of the building you want to block. If an unknown type is given the game will likely complain in loading screen (no check is done).
@param tileObject Table (optional) Tile object or table of tile object of the platform(s) you want to block the type outside of. If omitted, rule will be applied to all platform in default list.
@after data-updates
@before data-final-fixes
@alias block_all
]]--
buildingplatform.blockAll = function(buildingType,tileObject)
	buildingType = buildingType or "";
	tileObject = tileObject or nil;
	buildingplatform.checkParameters(buildingType,"")
	local tileObjectTable = buildingplatform.getTileObjectTable(tileObject)
	for _,tObject in pairs(tileObjectTable) do
		if(not tObject.buildingsAllowed[buildingType]) then --only block all if no rule of blocking already exist
			tObject.buildingsAllowed[buildingType] = {}; --empty table mean all building blocked
		end
	end
end

--[[API_DOC
@description Block all buildings of a given type which name does NOT match the pattern outside the given tileObject.

Calling this function will not remove previous permission.
It mean that you can allow a type on multiple tileObject by calling this function with the same type, but with each different tileObject you want.
@param buildingType string Factorio type of the building you want to block. If an unknown type is given the game will likely complain in loading screen (no check is done).
@param pattern string LUA pattern the name of an entity of the given type must match to be not blocked.

Patterns are checked by using LUA string.match function against the building name, so format it the right way as LUA pattern aren't regular expression like other languages.
@param tileObject Table (optional) Tile object or table of tile object of the platform(s) you want to block the type outside of. If omitted, rule will be applied to all platform in default list.
@after data-updates
@before data-final-fixes
@alias block_all_but
]]--
buildingplatform.blockAllBut = function(buildingType, pattern,tileObject)
	buildingType = buildingType or "";
	pattern = pattern or "";
	tileObject = tileObject or nil;
	buildingplatform.checkParameters(buildingType,pattern)
	local tileObjectTable = buildingplatform.getTileObjectTable(tileObject)
	for _,tObject in pairs(tileObjectTable) do
		if(not tObject.buildingsAllowed[buildingType]) then
			tObject.buildingsAllowed[buildingType] = {pattern,nil};
		else
			table.insert(tObject.buildingsAllowed[buildingType],pattern)
		end
	end
end

--[[API_DOC
@description Block all buildings of a given type which name match the pattern outside the given tileObject.

Calling this function will not remove previous permission.
It mean that you can allow a type on multiple tileObject by calling this function with the same type, but with each different tileObject you want.

Call to this function will have no effect if the given type have one or more block all (but) rules.
@param buildingType string Factorio type of the building you want to block. If an unknown type is given the game will likely complain in loading screen (no check is done).
@param pattern string LUA pattern the name of an entity of the given type must match to be blocked.

Patterns are checked by using LUA string.match function against the building name, so format it the right way as LUA pattern aren't regular expression like other languages.
@param tileObject Table (optional) Tile object or table of tile object of the platform(s) you want to block the type outside of. If omitted, rule will be applied to all platform in default list.
@after data-updates
@before data-final-fixes
@alias block_only
]]--
buildingplatform.blockOnly = function(buildingType, pattern,tileObject)
	buildingType = buildingType or "";
	pattern = pattern or "";
	tileObject = tileObject or nil;
	buildingplatform.checkParameters(buildingType,pattern)
	local tileObjectTable = buildingplatform.getTileObjectTable(tileObject)
	for _,tObject in pairs(tileObjectTable) do
		if(not tObject.buildingsDisallowed[buildingType]) then
			tObject.buildingsDisallowed[buildingType] = {pattern}
		else
			table.insert(tObject.buildingsDisallowed[buildingType],pattern)
		end
	end
end

--[[API_DOC
@description Disable own default rules the mod apply. It will not block other mod custom rules added by calling API functions.
Remember that people sometime want some buildings to be blocked and some other don't. Try to integrate the config file directives.

@after data-updates
@before data-final-fixes
@alias disable_building_platform_own_block_list
]]--
buildingplatform.disableBuildingPlatformOwnBlockList = function()
	buildingplatform.config.disableOwnBlocking = true;
end


--add alias with underscore instead of camelCase for those who prefer them.
buildingplatform.block_all = buildingplatform.blockAll
buildingplatform.block_all_but = buildingplatform.blockAllBut
buildingplatform.block_only = buildingplatform.blockOnly
buildingplatform.disable_building_platform_own_block_list = buildingplatform.disableBuildingPlatformOwnBlockList